# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import pytest


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--longtests",
        action="store_true",
        dest="longtests",
        default=False,
        help="enable tests marked as longtests",
    )


def pytest_configure(config: pytest.Config) -> None:
    """Register the 'longtests' marker to avoid warnings."""
    config.addinivalue_line("markers", "longtests: mark tests as long-running")


def pytest_collection_modifyitems(config: pytest.Config, items: list[pytest.Item]) -> None:
    """Automatically skip @pytest.mark.longtests tests unless --longtests is used."""
    if not config.getoption("--longtests"):
        skip_marker = pytest.mark.skip(reason="Skipping long test, use --longtests to enable")
        for item in items:
            if item.get_closest_marker("longtests"):
                item.add_marker(skip_marker)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

import numpy as np

from anemoi.inference.input import Input
from anemoi.inference.inputs import create_input
from anemoi.inference.types import Date
from anemoi.inference.types import State

from . import input_registry

LOG = logging.getLogger(__name__)


def _mask_and_combine_states(
    combined_state: State,
    new_state: State,
    combined_mask: Union[np.ndarray, slice],
    mask: np.ndarray,
    fields: Iterable[str],
) -> State:
    """Mask and combine two states.

    Parameters
    ----------
    combined_state : State
        The state to be combined into.
    new_state : State
        The other state to combine.
    combined_mask : Optional[np.ndarray]
        The mask to apply to combined_state. If None, no mask is applied
    mask : np.ndarray
        The mask to apply to new_state.
    fields: Iterable[str]
        The fields to combine in the states

    Returns
    -------
    State
        The combined state
    """
    for field in fields:
        combined_state[field] = np.concatenate(
            [combined_state[field][..., combined_mask], new_state[field][..., mask]],
            axis=-1,
        )

    return combined_state


@input_registry.register("cutout")
class Cutout(Input):
    """Combines one or more LAMs into a global source using cutouts."""

    def __init__(self, context, **sources: dict[str, dict]):
        """Create a cutout input from a list of sources.

        Parameters
        ----------
        context : dict
            The context runner.
        sources : dict of sources
            A dictionary of sources to combine.
        """
        super().__init__(context)

        self.sources: dict[str, Input] = {}
        self.masks: dict[str, np.ndarray] = {}
        for src, cfg in sources.items():
            if isinstance(cfg, str):
                mask = f"{src}/cutout_mask"
            else:
                mask = cfg.pop("mask", f"{src}/cutout_mask")
            self.sources[src] = create_input(context, cfg)
            self.masks[src] = self.sources[src].checkpoint.load_supporting_array(mask)

    def __repr__(self):
        return f"Cutout({self.sources})"

    def create_input_state(self, *, date: Optional[Date]) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """

        LOG.info(f"Concatenating states from {self.sources}")
        sources = list(self.sources.keys())

        combined_state = self.sources[sources[0]].create_input_state(date=date)
        combined_mask = self.masks[sources[0]]
        for source in sources[1:]:
            mask = self.masks[source]
            new_state = self.sources[source].create_input_state(date=date)

            combined_state = _mask_and_combine_states(
                combined_state, new_state, combined_mask, mask, ["longitudes", "latitudes"]
            )
            combined_state["fields"] = _mask_and_combine_states(
                combined_state["fields"], new_state["fields"], combined_mask, mask, combined_state["fields"]
            )
            combined_mask = slice(0, None)

        return combined_state

    def load_forcings_state(self, *, variables: List[str], dates: List[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            List of variables to load.
        dates : List[Date]
            List of dates for which to load the forcings.
        current_state : State
            The current state of the input.

        Returns
        -------
        State
            The loaded forcings state.
        """

        sources = list(self.sources.keys())
        combined_fields = self.sources[sources[0]].load_forcings_state(
            variables=variables, dates=dates, current_state=current_state
        )["fields"]
        combined_mask = self.masks[sources[0]]
        for source in sources[1:]:
            mask = self.masks[source]
            new_fields = self.sources[source].load_forcings_state(
                variables=variables, dates=dates, current_state=current_state
            )["fields"]
            combined_fields = _mask_and_combine_states(
                combined_fields, new_fields, combined_mask, mask, combined_fields
            )
            combined_mask = slice(0, None)

        current_state["fields"] |= combined_fields
        return current_state



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import Optional

from anemoi.inference.config import Configuration
from anemoi.inference.context import Context

from . import output_registry
from .masked import MaskedOutput

LOG = logging.getLogger(__name__)


@output_registry.register("apply_mask")
class ApplyMaskOutput(MaskedOutput):
    """Apply mask output class.

    Parameters
    ----------
    context : dict
        The context dictionary.
    mask : str
        The mask identifier.
    output : dict
        The output configuration dictionary.
    output_frequency : int, optional
        The frequency of output, by default None.
    write_initial_state : bool, optional
        Whether to write the initial state, by default None.
    """

    def __init__(
        self,
        context: Context,
        *,
        mask: str,
        output: Configuration,
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
    ) -> None:
        super().__init__(
            context,
            mask=context.checkpoint.load_supporting_array(mask),
            output=output,
            output_frequency=output_frequency,
            write_initial_state=write_initial_state,
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
import os
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

from earthkit.data.utils.dates import to_datetime
from omegaconf import DictConfig
from omegaconf import ListConfig
from omegaconf import OmegaConf
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import field_validator

LOG = logging.getLogger(__name__)

T = TypeVar("T", bound="Configuration")


class Configuration(BaseModel):
    """Configuration class."""

    model_config = ConfigDict(extra="forbid")

    date: Union[datetime, None] = None
    """The starting date for the forecast. If not provided, the date will depend on the selected Input object. If a string, it is parsed by :func:`earthkit.data.utils.dates`."""

    @field_validator("date", mode="before")
    @classmethod
    def to_datetime(cls, date: Union[str, int, datetime, None]) -> Optional[datetime]:
        if date is not None:
            return to_datetime(date)

    @classmethod
    def load(
        cls: Type[T],
        path: Union[str, Dict[str, Any]],
        overrides: Union[List[str], List[dict], str, dict] = [],
        defaults: Optional[Union[str, List[str], dict]] = None,
    ) -> T:
        """Load the configuration.

        Parameters
        ----------
        path : Union[str, dict]
            Path to the configuration file or a dictionary containing the configuration.
        overrides : Union[List[str], List[dict], str, dict], optional
            List of overrides to apply to the configuration, by default [].
        defaults : Optional[Union[str, List[str], dict]], optional
            Default values to set in the configuration, by default None.

        Returns
        -------
        Configuration
            The loaded configuration.
        """

        configs: List[Union[DictConfig, ListConfig]] = []

        # Set default values
        if defaults is not None:
            if not isinstance(defaults, list):
                defaults = [defaults]
            for d in defaults:
                if isinstance(d, str):
                    configs.append(OmegaConf.load(d))
                    continue
                configs.append(OmegaConf.create(d))

        # Load the user configuration
        if isinstance(path, dict):
            configs.append(OmegaConf.create(path))
        else:
            configs.append(OmegaConf.load(path))

        if not isinstance(overrides, list):
            overrides = [overrides]

        # unsafe merge should be fine as we don't re-use the original configs
        oc_config = OmegaConf.unsafe_merge(*configs)

        for override in overrides:
            if isinstance(override, dict):
                oc_config = OmegaConf.unsafe_merge(oc_config, OmegaConf.create(override))
            else:
                # use from_dotlist to use OmegaConf split
                # which allows for "param.val" or "param[val]".
                override_conf = OmegaConf.from_dotlist([override])
                # We can't directly merge reconstructed with the config because
                # omegaconf prefers parsing digits (like 0 in key.0) into dict keys
                # rather than lists.
                # Instead, we provide a reference config and we try to merge the override
                # into the reference and keep types provided by the reference.
                oc_config = OmegaConf.unsafe_merge(_merge_configs(oc_config, override_conf))

        resolved_config = OmegaConf.to_container(oc_config, resolve=True)

        config = cls.model_validate(resolved_config)

        # Set environment variables found in the configuration
        # as soon as possible
        for key, value in config.env.items():
            os.environ[key] = str(value)

        return config


def _merge_configs(ref_conf: Any, new_conf: Any) -> Any:
    """Recursively merges a new OmegaConf object into a reference OmegaConf object

    Parameters
    ----------
    ref_conf : Any
        reference OmegaConf object. Should be a DictConfig or ListConfig.
    new_conf : Any
        new OmegaConf object.

    Returns
    -------
    Any
        The merged OmegaConf config
    """
    if isinstance(new_conf, DictConfig) and len(new_conf):
        key, rest = next(iter(new_conf.items()))
        key = str(key)
    elif isinstance(new_conf, ListConfig) and len(new_conf):
        key, rest = 0, new_conf[0]
    else:
        return new_conf
    if isinstance(ref_conf, ListConfig):
        if isinstance(key, str) and not key.isdigit():
            raise ValueError(f"Expected int key, got {key}")
        index = int(key)
        if index < len(ref_conf):
            LOG.debug(f"key {key} is used as list key in list{ref_conf}")
            ref_conf[index] = _merge_configs(ref_conf[index], rest)
        elif index == len(ref_conf):
            LOG.debug(f"key {key} is used to append to list {ref_conf}")
            ref_conf.append(rest)
        else:
            raise IndexError(f"key {key} out of range for list {ref_conf} of length {len(ref_conf)}")
        return ref_conf
    elif isinstance(ref_conf, DictConfig) and key in ref_conf:
        ref_conf[key] = _merge_configs(ref_conf[key], rest)
        return ref_conf
    elif isinstance(ref_conf, DictConfig):
        ref_conf[key] = rest
        return ref_conf
    else:
        raise ValueError(f"ref is of unexpected type {type(ref_conf)}. Should be ListConfig or DictConfig")



import numpy as np

from anemoi.inference.inputs.cutout import _mask_and_combine_states


def test_mask_and_combine_states():
    states = [{"a": np.arange(5).astype(float)}, {"a": np.arange(5, 10).astype(float)}, {"a": np.arange(10, 15)}]
    masks = [np.zeros(5).astype(bool) for _ in states]
    masks[0][[1, 2]] = True
    masks[1][[2, 3]] = True
    masks[2][[2, 4]] = True
    combined_state = {k: states[0][k][:] for k in states[0]}
    combined_mask = masks[0]
    for k in range(1, len(states)):
        mask = masks[k]
        new_state = states[k]
        combined_state = _mask_and_combine_states(combined_state, new_state, combined_mask, mask, ["a"])
        combined_mask = slice(0, None)

    assert combined_state["a"].shape[0] == 6
    assert (
        combined_state["a"]
        == np.array(
            [
                states[0]["a"][1],
                states[0]["a"][2],
                states[1]["a"][2],
                states[1]["a"][3],
                states[2]["a"][2],
                states[2]["a"][4],
            ]
        )
    ).all()
